from typing import Optional
from jacobsjinjatoo import stringmanip

class TopicCreatorBase:
    def __init__(self, placeholder: str, root: Optional[str] = None):
        self._placeholder = placeholder
        if root is None:
            self._base_topic = None
        else:
            self._base_topic = root.strip("/")

    def slash(self, *args) -> str:
        if self._base_topic is None:
            return "/".join(args)
        else:
            return f"{self._base_topic}/{'/'.join(args)}"


class SignalTopicCreator(TopicCreatorBase):
    def __init__(self, placeholder: str, root: str):
        super().__init__(placeholder, root)

    def signal_topic(self, signal_name: str) -> str:
        return self.slash("signal", stringmanip.lower_camel_case(signal_name))


class MethodTopicCreator(TopicCreatorBase):
    def __init__(self, placeholder: str, root: str):
        super().__init__(placeholder, root)

    def method_topic(self, method_name: str) -> str:
        return self.slash("method", stringmanip.lower_camel_case(method_name))

    def method_response_topic(self, method_name: str, namespace: str, client_id: str) -> str:
        return f"client/{client_id}/{namespace}/methodResponse"


class PropertyTopicCreator(TopicCreatorBase):
    def __init__(self, placeholder: str, root: str):
        super().__init__(placeholder, root)

    def property_value_topic(self, property_name: str) -> str:
        return self.slash("property", stringmanip.lower_camel_case(property_name), "value")

    def property_update_topic(self, property_name: str) -> str:
        return self.slash("property", stringmanip.lower_camel_case(property_name), "setValue")

    def property_response_topic(self, property_name: str, namespace: str, client_id: str) -> str:
        return f"client/{client_id}/{namespace}/propertyUpdateResponse"

class InterfaceTopicCreator(TopicCreatorBase):
    """Helper class for creating MQTT topics for various stinger elements."""

    def __init__(self, interface_name: str, placeholder: str, root: Optional[str] = None):
        super().__init__(placeholder, root)
        self._interface_name = stringmanip.lower_camel_case(interface_name)

    @property
    def _topic_prefix(self) -> str:
        return self.slash(self._interface_name, self._placeholder)

    def interface_info_topic(self) -> str:
        return f"{self._topic_prefix}/interface"

    def signal_topic_creator(self) -> SignalTopicCreator:
        return SignalTopicCreator(self._placeholder, self._topic_prefix)

    def method_topic_creator(self) -> MethodTopicCreator:
        return MethodTopicCreator(self._placeholder, self._topic_prefix)

    def property_topic_creator(self) -> PropertyTopicCreator:
        return PropertyTopicCreator(self._placeholder, self._topic_prefix)



import unittest
from stingeripc.components import Arg, ArgPrimitive, InvalidStingerStructure
from stingeripc.args import ArgPrimitiveType

class TestArgCreationFromSpec(unittest.TestCase):

    def setUp(self):
        self.spec_objs = [
            {"name": "one", "type": "integer", "python_type": "int", "value_type": ArgPrimitiveType.INTEGER},
            {"name": "two", "type": "float", "python_type": "float", "value_type": ArgPrimitiveType.FLOAT},
            {"name": "three", "type": "boolean", "python_type": "bool", "value_type": ArgPrimitiveType.BOOLEAN},
            {"name": "four", "type": "string", "python_type": "str", "value_type": ArgPrimitiveType.STRING},
        ]

    def test_arg(self):
        for obj in self.spec_objs:
            with self.subTest(obj):
                arg = Arg.new_arg_from_stinger(obj)
                self.assertEqual(arg.name, obj['name'])
                self.assertEqual(arg.python_type, obj['python_type'])
                self.assertEqual(arg.type, obj['value_type'])
                example = arg.get_random_example_value()
                real_python_type = eval(arg.python_type)
                self.assertIsInstance(example, real_python_type)


class TestInvalidArgCreation(unittest.TestCase):

    def test_not_an_arg_type_from_spec(self):
        with self.assertRaises(InvalidStingerStructure):
            ArgPrimitive.new_arg_primitive_from_stinger({"name": "one", "type": "double"})

    def test_not_an_arg_type(self):
        with self.assertRaises(InvalidStingerStructure):
            ArgPrimitiveType.from_string("unsigned int")

    def test_arg_primitive_name_missing(self):
        with self.assertRaises(InvalidStingerStructure):
            ArgPrimitive.new_arg_primitive_from_stinger({"type": "integer"})  

    def test_arg_name_missing(self):
        with self.assertRaises(InvalidStingerStructure):
            Arg.new_arg_primitive_from_stinger({"type": "integer"})  
    
    def test_arg_primitive_type_missing(self):
        with self.assertRaises(InvalidStingerStructure):
            ArgPrimitive.new_arg_primitive_from_stinger({"name": "foo"})  

    def test_arg_type_missing(self):
        with self.assertRaises(InvalidStingerStructure):
            Arg.new_arg_primitive_from_stinger({"name": "foo"})  

class TestPythonTypes(unittest.TestCase):

    def test_from_invalid_input(self):
        with self.assertRaises(InvalidStingerStructure):
            ArgPrimitiveType.to_python_type(-1)


from stingeripc.topic import InterfaceTopicCreator

import unittest


class TestTopics(unittest.TestCase):
    def test_signal_topic(self):
        itc = InterfaceTopicCreator("test_interface")
        stc = itc.signal_topic_creator()
        self.assertEqual(stc.signal_topic("alert"), "test_interface/signal/alert")



"""
An `Arg` is used in one of the following contexts:
 * An argument to a method
 * A component of the return type of a method
 * A component of a signal
 * A component of a property

An `Arg` can be defined in several different ways:
 * `Primitive` means that the type of the Arg is a primitive type that is directly provided.
 * `Enum` means that the type of the Arg is an integer and the value of the integer is controlled by an enumerated list.
 * `Struct` means that the type of the Arg is a structure make up of additional args.

When the `Arg` type is `Primitive` then it has an `ArgPrimitiveType` which controlls which type the arg is.
"""

from __future__ import annotations
from enum import Enum
from .exceptions import InvalidStingerStructure


class ArgType(Enum):
    UNKNOWN = 0
    PRIMITIVE = 1
    ENUM = 2
    STRUCT = 3
    DATETIME = 4
    DURATION = 5
    BINARY = 6
    ARRAY = 7


class ArgPrimitiveType(Enum):
    """This is not an Arg, rather it is an enumeration of different primitives that an ArgPrimitive can represent."""

    BOOLEAN = 0
    INTEGER = 1
    FLOAT = 2
    STRING = 3

    @classmethod
    def from_string(cls, arg_type: str) -> ArgPrimitiveType:
        if hasattr(cls, arg_type.upper()):
            return getattr(cls, arg_type.upper())
        else:
            raise InvalidStingerStructure(f"No ArgType called '{arg_type}'")

    @classmethod
    def to_python_type(cls, arg_type: ArgPrimitiveType, optional: bool = False) -> str:
        if optional:
            return f"Optional[{cls.to_python_type(arg_type, optional=False)}]"
        if arg_type == cls.BOOLEAN:
            return "bool"
        elif arg_type == cls.INTEGER:
            return "int"
        elif arg_type == cls.FLOAT:
            return "float"
        elif arg_type == cls.STRING:
            return "str"
        raise InvalidStingerStructure("Unhandled arg type")

    @classmethod
    def to_rust_type(cls, arg_type: ArgPrimitiveType, optional: bool = False) -> str:
        if optional:
            return f"Option<{cls.to_rust_type(arg_type, optional=False)}>"
        if arg_type == cls.BOOLEAN:
            return "bool"
        elif arg_type == cls.INTEGER:
            return "i32"
        elif arg_type == cls.FLOAT:
            return "f32"
        elif arg_type == cls.STRING:
            return "String"
        raise InvalidStingerStructure("Unhandled arg type")

    @classmethod
    def to_json_type(cls, arg_type: ArgPrimitiveType) -> str:
        if arg_type == cls.BOOLEAN:
            return "boolean"
        elif arg_type == cls.INTEGER:
            return "integer"
        elif arg_type == cls.FLOAT:
            return "number"
        elif arg_type == cls.STRING:
            return "string"
        raise InvalidStingerStructure("Unhandled arg type")

    @classmethod
    def to_cpp_type(cls, arg_type: ArgPrimitiveType, optional: bool = False) -> str:
        if arg_type == cls.BOOLEAN:
            if optional:
                return "std::optional<bool>"
            else:
                return "bool"
        elif arg_type == cls.INTEGER:
            if optional:
                return "std::optional<int>"
            else:
                return "int"
        elif arg_type == cls.FLOAT:
            if optional:
                return "std::optional<double>"
            else:
                return "double"
        elif arg_type == cls.STRING:
            if optional:
                return "std::optional<std::string>"
            else:
                return "std::string"
        raise InvalidStingerStructure("Unhandled arg type")

    @classmethod
    def to_cpp_rapidjson_type_str(cls, arg_type: ArgPrimitiveType) -> str:
        if arg_type == cls.BOOLEAN:
            return "Bool"
        elif arg_type == cls.INTEGER:
            return "Int"
        elif arg_type == cls.FLOAT:
            return "Double"
        elif arg_type == cls.STRING:
            return "String"
        raise InvalidStingerStructure("Unhandled arg type")




import unittest
import yaml
import yamlloader
import os.path
import jsonschema_rs


example_stinger_path = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "../example_interfaces"))
schema_path = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "../schemas/schema.yaml"))


class SchemaValidation:

    def test_validates(self):
        with open(schema_path) as fp:
            schema_obj = yaml.load(fp, Loader=yamlloader.ordereddict.Loader)
        validator = jsonschema_rs.validator_for(schema_obj)
        
        param_only_path = os.path.join(example_stinger_path, self.directory, self.stinger_name)
        with open(param_only_path) as fp:
            interface = yaml.load(fp, Loader=yamlloader.ordereddict.Loader)
        
        errors = list(validator.iter_errors(interface))
        if errors:
            for error in errors:
                print(f"Error: {error}")
                print(f"Location: {error.instance_path}")
            assert (len(errors), 0)

class TestParamOnlyValidatesAgainstSchema(unittest.TestCase, SchemaValidation):

    def setUp(self):
        self.directory = "signal_only"
        self.stinger_name = "signal_only.stingeripc"


class TestEnumOnlyValidatesAgainstSchema(unittest.TestCase, SchemaValidation):

    def setUp(self):
        self.directory = "enum_only"
        self.stinger_name = "enum_only.stingeripc"




from typing import Optional, Any, Dict, IO, Union
import yaml
import yamlloader

from .components import StingerSpec, InvalidStingerStructure
from .topic import InterfaceTopicCreator

VERSION_SUPPORTED = "0.0.7"


class StingerInterface(StingerSpec):

    def __init__(self, stinger: Dict[str, Any], topic_prefix: Optional[str] = None):
        itc = self._create_topic_creator(stinger)
        super().__init__(itc, stinger["interface"])

    @staticmethod
    def _create_topic_creator(
        stinger_spec: Dict[str, Any], placeholder: str='{}', topic_prefix: Optional[str] = None
    ) -> InterfaceTopicCreator:
        if (
            "stingeripc" not in stinger_spec
            or "version" not in stinger_spec["stingeripc"]
            or stinger_spec["stingeripc"]["version"] not in [VERSION_SUPPORTED]
        ):
            raise InvalidStingerStructure(
                f"Provided Stinger Spec does not claim to be version {VERSION_SUPPORTED}"
            )
        if "interface" not in stinger_spec or "name" not in stinger_spec["interface"]:
            raise InvalidStingerStructure(
                "Could not find interface name in Stinger Spec"
            )
        itc = InterfaceTopicCreator(
            stinger_spec["interface"]["name"], placeholder, root=topic_prefix
        )
        return itc

    @classmethod
    def from_yaml(cls, yaml_input: Union[str, IO], placeholder:str='{}') -> StingerSpec:
        yaml_obj = yaml.load(yaml_input, Loader=yamlloader.ordereddict.Loader)
        itc = cls._create_topic_creator(yaml_obj, placeholder)
        return cls.new_spec_from_stinger(itc, yaml_obj)

    @classmethod
    def from_dict(cls, stinger_dict: Dict[str, Any], placeholder:str='{}') -> StingerSpec:
        itc = cls._create_topic_creator(stinger_dict, placeholder)
        return cls.new_spec_from_stinger(itc, stinger_dict)


from stingeripc.components import InterfaceEnum

import unittest

class TestEnumCreateManually(unittest.TestCase):
    def setUp(self):
        self.enum = InterfaceEnum("bread")
        self.enum.add_value("pumpernickel")
        self.enum.add_value("whole wheat")
        self.enum.add_value("french")

    def test_name(self):
        self.assertEqual(self.enum.name, "bread")
    
    def test_values(self):
        self.assertEqual(len(self.enum.values), 3)


class TestSignalCreateFromStinger(unittest.TestCase):
    def setUp(self):
        stinger = [
            {"name": "white"},
            {"name": "blue"},
            {"name": "red"},
        ]
        self.enum = InterfaceEnum.new_enum_from_stinger("color", stinger)

    def test_name(self):
        self.assertEqual(self.enum.name, "color")
    
    def test_values(self):
        self.assertEqual(len(self.enum.values), 3)




"""
For both functions, the interface provided is the raw dictionary parsed from the Stinger YAML file, as specified in `schema/schema.yaml`.

"""
import semantic_version
from typing import Dict, Any, Set
from copy import deepcopy
from stingeripc.components import StingerSpec

def _collect_used_types(args: list[Dict[str, Any]], used_enums: Set[str], used_structs: Set[str]) -> None:
    """Recursively collect enum and struct names used in argument lists."""
    for arg in args:
        arg_type = arg.get("type")
        if arg_type == "enum":
            used_enums.add(arg["enumName"])
        elif arg_type == "struct":
            used_structs.add(arg["structName"])
        elif arg_type == "array":
            # Check the item type of arrays
            item_type = arg.get("itemType", {})
            if isinstance(item_type, dict):
                _collect_used_types([item_type], used_enums, used_structs)


def _collect_struct_dependencies(struct_spec: Dict[str, Any], used_enums: Set[str], used_structs: Set[str]) -> None:
    """Recursively collect enums and structs used by a struct's members."""
    members = struct_spec.get("members", [])
    for member in members:
        member_type = member.get("type")
        if member_type == "enum":
            used_enums.add(member["enumName"])
        elif member_type == "struct":
            struct_name = member["structName"]
            # Avoid infinite recursion if structs reference each other
            if struct_name not in used_structs:
                used_structs.add(struct_name)
        elif member_type == "array":
            item_type = member.get("itemType", {})
            if isinstance(item_type, dict):
                _collect_used_types([item_type], used_enums, used_structs)


def filter_by_consumer(interface: StingerSpec, consumer_name: str) -> Dict[str, Any]:
    """
    Filter a StingerInterface to only include methods, properties, signals, and events that specify the given consumer.

    Args:
        interface: The original StingerInterface.
        consumer_name: The name of the consumer to filter by.
        
    Returns:
        A new StingerInterface containing only the elements that are required for a client interface for the given consumer.

    Methodology:
        - Methods are included if they have the consumer in their consumers list.
        - Properties are included if they have the consumer in their consumers list.
        - Signals are included if they have the consumer in their consumers list

    Enums are included if they are used by any of the included methods, properties, signals, or by any structs used by those elements.

    Enums are included if they are used by any of the included methods, properties, signals, or by any structs used by those elements.
    """
    # Create a deep copy to avoid modifying the original
    filtered = deepcopy(interface)
    
    # Track which enums and structs are used
    used_enums: Set[str] = set()
    used_structs: Set[str] = set()
    
    # Filter signals
    if "signals" in filtered:
        filtered_signals = {}
        for signal_name, signal_spec in filtered["signals"].items():
            consumers = signal_spec.get("consumers", [])
            # If no consumers specified, include for all consumers
            # If consumers specified, only include if consumer_name is in the list
            if consumer_name in consumers:
                filtered_signals[signal_name] = signal_spec
                # Collect types used in payload
                payload = signal_spec.get("payload", [])
                _collect_used_types(payload, used_enums, used_structs)
        filtered["signals"] = filtered_signals
    
    # Filter properties
    if "properties" in filtered:
        filtered_properties = {}
        for prop_name, prop_spec in filtered["properties"].items():
            consumers = prop_spec.get("consumers", [])
            if consumer_name in consumers:
                filtered_properties[prop_name] = prop_spec
                # Collect types used in values
                values = prop_spec.get("values", [])
                _collect_used_types(values, used_enums, used_structs)
        filtered["properties"] = filtered_properties
    
    # Filter methods
    if "methods" in filtered:
        filtered_methods = {}
        for method_name, method_spec in filtered["methods"].items():
            consumers = method_spec.get("consumers", [])
            if consumer_name in consumers:
                filtered_methods[method_name] = method_spec
                # Collect types used in arguments
                arguments = method_spec.get("arguments", [])
                _collect_used_types(arguments, used_enums, used_structs)
                # Collect types used in return values
                return_values = method_spec.get("returnValues", [])
                _collect_used_types(return_values, used_enums, used_structs)
        filtered["methods"] = filtered_methods
    
    # Now collect dependencies from used structs
    if "structures" in interface:
        # Keep iterating until no new structs are added
        prev_struct_count = -1
        while len(used_structs) != prev_struct_count:
            prev_struct_count = len(used_structs)
            for struct_name in list(used_structs):
                if struct_name in interface["structures"]:
                    _collect_struct_dependencies(
                        interface["structures"][struct_name], 
                        used_enums, 
                        used_structs
                    )
    
    # Filter enums to only include used ones
    if "enums" in filtered:
        filtered_enums = {
            enum_name: enum_spec
            for enum_name, enum_spec in filtered["enums"].items()
            if enum_name in used_enums
        }
        filtered["enums"] = filtered_enums
    
    # Filter structs to only include used ones
    if "structures" in filtered:
        filtered_structs = {
            struct_name: struct_spec
            for struct_name, struct_spec in filtered["structures"].items()
            if struct_name in used_structs
        }
        filtered["structures"] = filtered_structs
    
    return filtered

def check_version_consistency(interface: Dict[str, Any]) -> None:
    """
    Anywhere there is a structVersion or enumVersion specified in an argument list, make sure that there is a version specified on the corresponding struct or enum.
    Also make sure the two versions are compatible, which means the major and minor versions must match.

    Raises:
        ValueError: If any method, property, signal, or event specifies consumers but does not specify a version.
    """
    def check_args(args: list[Dict[str, Any]], component_name: str, component_type: str) -> None:
        """Check version consistency for a list of arguments."""
        for arg in args:
            arg_type = arg.get("type")
            
            # Check enum version consistency
            if arg_type == "enum":
                enum_name = arg.get("enumName")
                enum_version_in_arg = arg.get("enumVersion")
                
                if enum_version_in_arg:
                    # Get the enum definition
                    if "enums" not in interface or enum_name not in interface["enums"]:
                        raise ValueError(
                            f"{component_type} '{component_name}' references enum '{enum_name}' "
                            f"which is not defined in the interface"
                        )
                    
                    enum_spec = interface["enums"][enum_name]
                    enum_version_in_def = enum_spec.get("version")
                    
                    if not enum_version_in_def:
                        raise ValueError(
                            f"{component_type} '{component_name}' specifies enumVersion '{enum_version_in_arg}' "
                            f"for enum '{enum_name}', but the enum definition does not have a version"
                        )
                    
                    # Check semantic version compatibility (major.minor must match)
                    arg_ver = semantic_version.Version(enum_version_in_arg)
                    def_ver = semantic_version.Version(enum_version_in_def)
                    
                    if arg_ver.major != def_ver.major or arg_ver.minor != def_ver.minor:
                        raise ValueError(
                            f"{component_type} '{component_name}' requires enum '{enum_name}' "
                            f"version {enum_version_in_arg} (major.minor: {arg_ver.major}.{arg_ver.minor}), "
                            f"but the enum is defined with version {enum_version_in_def} "
                            f"(major.minor: {def_ver.major}.{def_ver.minor})"
                        )
            
            # Check struct version consistency
            elif arg_type == "struct":
                struct_name = arg.get("structName")
                struct_version_in_arg = arg.get("structVersion")
                
                if struct_version_in_arg:
                    # Get the struct definition
                    if "structures" not in interface or struct_name not in interface["structures"]:
                        raise ValueError(
                            f"{component_type} '{component_name}' references struct '{struct_name}' "
                            f"which is not defined in the interface"
                        )
                    
                    struct_spec = interface["structures"][struct_name]
                    struct_version_in_def = struct_spec.get("version")
                    
                    if not struct_version_in_def:
                        raise ValueError(
                            f"{component_type} '{component_name}' specifies structVersion '{struct_version_in_arg}' "
                            f"for struct '{struct_name}', but the struct definition does not have a version"
                        )
                    
                    # Check semantic version compatibility (major.minor must match)
                    arg_ver = semantic_version.Version(struct_version_in_arg)
                    def_ver = semantic_version.Version(struct_version_in_def)
                    
                    if arg_ver.major != def_ver.major or arg_ver.minor != def_ver.minor:
                        raise ValueError(
                            f"{component_type} '{component_name}' requires struct '{struct_name}' "
                            f"version {struct_version_in_arg} (major.minor: {arg_ver.major}.{arg_ver.minor}), "
                            f"but the struct is defined with version {struct_version_in_def} "
                            f"(major.minor: {def_ver.major}.{def_ver.minor})"
                        )
            
            # Recursively check array item types
            elif arg_type == "array":
                item_type = arg.get("itemType")
                if isinstance(item_type, dict):
                    check_args([item_type], component_name, component_type)
    
    # Check signals
    if "signals" in interface:
        for signal_name, signal_spec in interface["signals"].items():
            # Check if consumers specified but version is missing
            if signal_spec.get("consumers") and not signal_spec.get("version"):
                raise ValueError(
                    f"Signal '{signal_name}' specifies consumers but does not specify a version"
                )
            
            # Check payload arguments
            payload = signal_spec.get("payload", [])
            check_args(payload, signal_name, "Signal")
    
    # Check properties
    if "properties" in interface:
        for prop_name, prop_spec in interface["properties"].items():
            # Check if consumers specified but version is missing
            if prop_spec.get("consumers") and not prop_spec.get("version"):
                raise ValueError(
                    f"Property '{prop_name}' specifies consumers but does not specify a version"
                )
            
            # Check values arguments
            values = prop_spec.get("values", [])
            check_args(values, prop_name, "Property")
    
    # Check methods
    if "methods" in interface:
        for method_name, method_spec in interface["methods"].items():
            # Check if consumers specified but version is missing
            if method_spec.get("consumers") and not method_spec.get("version"):
                raise ValueError(
                    f"Method '{method_name}' specifies consumers but does not specify a version"
                )
            
            # Check arguments
            arguments = method_spec.get("arguments", [])
            check_args(arguments, method_name, "Method")
            
            # Check return values
            return_values = method_spec.get("returnValues", [])
            check_args(return_values, method_name, "Method")
    
    # Check struct members recursively
    if "structures" in interface:
        for struct_name, struct_spec in interface["structures"].items():
            members = struct_spec.get("members", [])
            check_args(members, struct_name, "Struct")



from jacobsjinjatoo import templator as jj2
import os
import shutil
import typer
from typing_extensions import Annotated
from typing import Optional, Any
from pathlib import Path
from rich import print
import importlib.resources

import yaml

from stingeripc import StingerInterface
from stingeripc.filtering import filter_by_consumer
import tomllib
import logging
import yaml
import yamlloader

#logging.basicConfig(level=logging.INFO)

def main(
    inname: Annotated[Path, typer.Argument(exists=True, file_okay=True, dir_okay=False, readable=True)],
    outdir: Annotated[Path, typer.Argument(file_okay=False, dir_okay=True, writable=True, readable=True)],
    language: Annotated[Optional[str], typer.Argument(help="Shortcut for internally provided templates")] = None,
    template_pkg: Annotated[Optional[list[str]], typer.Option(help="Python package(s) containing templates")] = None,
    template_path: Annotated[Optional[list[Path]], typer.Option(help="Filesystem path(s) to template directories")] = None,
    consumer: Annotated[Optional[str], typer.Option("--consumer", help="Consumer name/identifier")] = None,
    config: Annotated[Optional[Path], typer.Option("--config", help="TOML configuration file", exists=True, file_okay=True, dir_okay=False, readable=True)] = None,
):
    """Generate output for a Stinger interface.
    
    At least one of --language, --template-pkg, or --template-path must be provided.
    """
    
    # Validate that at least one template source is provided
    if not language and not template_pkg and not template_path:
        raise typer.BadParameter(
            "At least one of: --language, --template-pkg, or --template-path must be provided"
        )

    config_obj = None
    if config:
        with config.open("rb") as f:
            config_obj = tomllib.load(f)

    if consumer:
        print(f"💠 CONSUMER {consumer}")
        with inname.open(mode="r") as f:
            yaml_obj = yaml.load(f, Loader=yamlloader.ordereddict.Loader)
            stinger_yaml = filter_by_consumer(yaml_obj, consumer)
            stinger = StingerInterface.from_dict(stinger_yaml)
    else:
        with inname.open(mode="r") as f:
            stinger = StingerInterface.from_yaml(f)

    params: dict[str, Any] = {"stinger": stinger, "config": config_obj}

    if outdir.is_file():
        raise RuntimeError("Output directory is a file!")
    
    print(f"📁 [bold cyan]OUTPUT:[/bold cyan] {outdir}")
    
    if not outdir.is_dir():
        print(f"📁  [green]MKDIR[/green]: {outdir}")
        os.makedirs(outdir)

    # Collect all template directories
    template_dirs = []
    
    code_templator = jj2.CodeTemplator(output_dir=outdir)
    web_templator = jj2.WebTemplator(output_dir=outdir)

    # Add template paths
    if template_path:
        for path in template_path:
            if not path.exists():
                raise RuntimeError(f"Template path does not exist: {path}")
            if not path.is_dir():
                raise RuntimeError(f"Template path is not a directory: {path}")
            real_path = path.resolve()
            print(f"[bold cyan]TEMPLATES:[/bold cyan] {real_path}")
            code_templator.add_template_dir(real_path)
            template_dirs.append(real_path)

    # Add template packages
    if template_pkg:
        for pkg_name in template_pkg:
            try:
                print(f"[bold cyan]TEMPLATES:[/bold cyan] {pkg_name}")
                code_templator.add_template_package(pkg_name)
                pkg_path = importlib.resources.files(pkg_name)
                template_dirs.append(Path(pkg_path))
            except ModuleNotFoundError as e:
                raise RuntimeError(f"Template package not found: {pkg_name}") from e
    
    # Add language-based template directory if language is specified
    if language:
        this_file = Path(__file__)
        this_dir = this_file.parent
        template_dir = (this_dir / "../templates" / language).resolve()
        print(f"[bold cyan]TEMPLATES:[/bold cyan] {template_dir}")
        if not template_dir.exists():
            raise RuntimeError(f"Template directory does not exist: {template_dir}")
        code_templator.add_template_dir(template_dir)
        template_dirs.append(template_dir)

    def recursive_find_output_files(src_walker: Path, dest_walker: Path) -> list[str]:
        found_files = []
        for entry in os.listdir(src_walker):
            src_entry = src_walker / entry
            dest_entry = dest_walker / entry
            if '{{' in entry and '}}' in entry:
                rendered_entry_name = code_templator.render_string(entry, **params)
                dest_entry = dest_walker / rendered_entry_name
            if entry.endswith(".jinja2"):
                dest_path_str = str(dest_entry)[:-len(".jinja2")]
                dest_path = Path(dest_path_str).relative_to(outdir)
                found_files.append(str(dest_path))
            elif src_entry.is_dir():
                found_files.extend(recursive_find_output_files(src_entry, dest_entry))
            elif src_entry.is_file():
                dest_path = Path(dest_entry).relative_to(outdir)
                found_files.append(str(dest_path))
        return found_files
    
    output_file_list = set()
    for template_dir in template_dirs:
        output_file_list.update(recursive_find_output_files(Path(template_dir), Path(outdir)))
    params["output_files"] = list(output_file_list)

    def recursive_render_templates(template_dir, src_walker: Path, dest_walker: Path):
        print(f"🚶   [green]WALK[/green]: {src_walker}")
        for entry in os.listdir(src_walker):
            src_entry = src_walker / entry
            dest_entry = dest_walker / entry
            print(f"🚶  [white]ENTRY[/white]: {src_entry.relative_to(template_dir)}")
            if '{{' in entry and '}}' in entry:
                rendered_entry_name = code_templator.render_string(entry, **params)
                dest_entry = dest_walker / rendered_entry_name
                print(f"👓   [grey]NAME[/grey]: {entry} -> {rendered_entry_name}")
            if entry.endswith(".jinja2"):
                dest_path_str = str(dest_entry)[:-len(".jinja2")]
                template = src_entry.relative_to(template_dir)
                dest_path = Path(dest_path_str).relative_to(outdir)
                print(f"✨  [green]GENER[/green]: {dest_path}")
                if dest_path_str.endswith(".html") or dest_path_str.endswith(".htm"):
                    web_templator.render_template(template, dest_path, **params)
                else:
                    code_templator.render_template(template, dest_path, **params)
            elif src_entry.is_dir():
                print(f"📁  [green]MKDIR[/green]: {dest_entry.resolve()}")
                if not dest_entry.exists():
                    dest_entry.mkdir(parents=True)
                recursive_render_templates(template_dir, src_entry, dest_entry)
            elif src_entry.is_file():
                shutil.copyfile(src_entry, dest_entry)
                print(f"📄   [green]COPY[/green]: {src_entry}")
            else:
                print(f"⚠️    [red]SKIP[/red]: {src_entry} (unknown type)")

    # Process templates from all template directories
    for template_dir in template_dirs:
        src = Path(template_dir)
        recursive_render_templates(template_dir, src, Path(outdir))

def run():
    typer.run(main)


if __name__ == "__main__":
    run()